/*[a-,body+,h-,o=100,r+,rec+,t=4,u+,#+,j=20/57/1$,n+]*/
/* UMemory.p */
/* Copyright  1985-1989 by Apple Computer, Inc.  All rights reserved. */

/*[f-]*/
/*
	This unit implements MacApp's memory management and segment management
	schemes.

	The memory management scheme works by distinguishing between "permanent"
	and "temporary" heap allocation requests.	Permanent requests are
	typically used for data your application allocates:  objects and
	handles.  Temporary requests are used for code segments and Toolbox
	resources and data, such as WDEF's and CDEF's.

	Permanent memory objects are considered permanent because they
	are not purged from memory until you explicitly dispose of or free them.
	(Of course, the Macintosh Memory Manager will purge them if they're
	marked purgeable.)	Permanent objects are allocated with NewPermHandle.
	This prevents the MacApp GrowZone from purging temporary objects to
	accommodate a permanent one.  In MacApp, all TObjects and its descendants
	are considered permanent, and are allocated with NewPermHandle.

	Temporary memory objects are considered temporary because MacApp's
	GrowZone procedure may purge them from memory to satisfy a memory
	allocation request.  This is true regardless of whether the object is
	marked non-purgeable, although the GrowZone procedure will not purge
	locked objects (such as code segments in use).	Typically, temporary
	objects are marked non-purgeable so that MacApp's GrowZone can control
	when they are purged.

	MacApp reserves a specific amount of heap space for temporary objects,
	the idea being that the space reserved is large enough to handle the
	largest number of temporary objects (e.g. code and system resources)
	needed at any given time.	If this amount is sufficiently large, your
	program will never fail loading a segment or system resource.	This
	amount is defined by the internal variable gSzCodeReserve.	You can
	retrieve its value by calling GetReserveSize.	It is initially set by
	the 'seg!' and 'mem!' resources, and can be changed at run-time by
	calling SetReserveSize.  The procedure BuildCodeReserve reserves the
	space by allocating the handle gCodeReserve and setting its size to
	gSzCodeReserve (the total size of all temporary objects in memory).

	When a temporary object is loaded into memory, the size of gCodeReserve
	is adjusted accordingly. Permanent objects can be loaded into memory
	only so long as there will still be at least gSzCodeReserve bytes
	available for temporary objects.

	To identify which objects in the heap are temporary, MacApp maintains
	four lists of handles.	The objects identified by these handles are
	considered temporary and fall under the control of MacApp's GrowZone
	procedure.	The lists are:

	gCodeSegs - A list of handles to all CODE resources in the application
	and system resource forks.

	gSysMemList - A list handles to RAM-based system resources.
	By default this list includes all PACK, LDEF, MDEF, CDEF and WDEF
	resources in the system and application resource forks. You can add
	other resources, such as fonts, by calling AddHandle.

	gApp1MemList, gApp2MemList - Lists of handles to application data or
	resources.	MacApp initializes both lists to NIL.  The difference
	between the two lists is that handles in gApp1MemList are purged
	before those in gApp2MemList.  You may add your own handles to these
	lists by allocating them with NewHandle and calling AddHandle for each
	handle to be added to the list.

	The Macintosh Memory Manager calls MacApp's GrowZone procedure only when
	all purgeable objects have been purged from the heap and there is still
	insufficient space to satisfy a memory request.  The GrowZone will
	look through the lists of temporary objects in memory, making one
	purgeable so long as there is still at least gSzCodeReserve bytes
	allocated to temporary memory (the total size of the temporary objects
	in memory and the gCodeReserve handle).  The GrowZone procedure attempts
	to never allow the size of the temporary objects and reserve to fall
	below gSzCodeReserve, thereby guaranteeing that space is always avail-
	able for code segments and system resources (provided gSzCodeReserve is
	large enough to handle your application at its most memory intensive
	state).

	The value of gSzCodeReserve is determined at startup-time by adding up
	the size of all the segments named in the 'seg!' resources, and adding
	the first value of each of the 'mem!' resources.  You can derive these
	resources by observing your program and using the MacApp debugger to
	help you determine when your application uses the largest amount of
	code and system resources.	Typically, we've found that this happens
	while printing on the LaserWriter, or during initialization or term-
	ination.

	MacApp maintains another reserve, known as the low memory reserve.	This
	is a kind of emergency reserve--when all else fails we release the
	gMemReserve handle.  Its size is initially determined by the second
	number of each 'mem!' resource, and can be changed by calling
	SetReserveSize.  You can retrieve its size by calling GetReserveSize.
	Internally, the low-memory reserve is allocated with the gMemReserve
	handle, and its size is stored in gSzMemReserve.

	The function InitUMemory is responsible for initially setting up the
	temporary and low-memory reserves, setting up the GrowZone procedure,
	initializing handle lists, and patching LoadSeg.  It returns true if
	the temporary reserve could be allocated.
*/
/*[f+]*/
#ifndef  __UMemory__
#define __UMemory__  0
#endif
#if  ! __UMemory__
#define __UMemory__  1

		/*  Auto-Include the requirements for this unit's interface. */
#ifndef  __TYPES__
#include "Types.h"
#endif
#ifndef  __MEMORY__
#include "Memory.h"
#endif
#if  ! qDebug

const short gUnloadAllSegs		= true;					/* Always TRUE when not debugging */
#endif

typedef Handle AHandleList[5000];
typedef AHandleList *PHandleList;
typedef PHandleList *HHandleList;							 /* A list of handles */

extern pascal long gMaxLockedRsrc;
#if  qDebug
extern pascal Boolean gMemMgtBreak;						/* if TRUE, break into debugger rather than
                                                         just report memory mgt. information */
extern pascal Boolean gRsrcReport;						/* report resource maximums */
extern pascal Boolean gSegReport;						/* report segment loadings */
#endif
extern pascal HHandleList gSysMemList;					/* List of system handles used to compute
                                                         current allocated temporary memory (all
                                                         PACK, LDEF, MDEF, CDEF, WDEF resources) */
extern pascal HHandleList gApp1MemList;
extern pascal HHandleList gApp2MemList;
			/* These start out NIL. You can create lists of handles and place
            then in either of these variables. (One list might be permanent
            handles in your application, and the other based on the current
            situation.) The values stored here should be a handle to a
            list of other handles.
            
            If you modify either of these lists, call CheckReserve. It
            calls BuildAllReserves to recompute the code and low space
            reserve.  CheckReserve's result indicates whether the full code
            reserve is present. If not, then your program may crash because
            a segment (or defproc) can't be loaded.
            
            The handles in the lists should normally be resources that your
            application might require at any given time.  All these handles
            will normally be marked non-purgeable. The GrowZone proc,
            however, can purge any of these handles that are not locked.*/

extern pascal HHandleList gCodeSegs;					/* List of all code segments */
extern pascal HHandleList gNonResidents;				/* List of non resident segments; a copy of
                                                         gCodeSegs with resident segments replaced
                                                         by NIL */
#if  qDebug
extern pascal Boolean gUnloadAllSegs;					/* UnloadAllSegments doesn't do anything if
                                                         this is false. */
#endif

			/* I N I T I A L I Z A T I O N */

extern pascal void InitUMemory(void);
		/* S E G M E N T   L O A D I N G */

		/* DEBUGGING NOTE: You cannot set a MacApp breakpoint at any of these
        routines, because they must not call anything (eg. %_BP) that may
        require a segment load. */

extern pascal short GetSegNumber(Ptr aProc);

extern pascal Size GetSegSize(short segnum);

extern pascal long LoadMacAppSegment(short segnum);

extern pascal void LoadResidentSegments(void);

extern pascal Boolean PreloadSegment(short segnum);

extern pascal void SetResidentSegment(short segnum, Boolean makeResident);

extern pascal void UnloadAllSegments(void);

		/* H A N D L E L I S T   M A N A G E M E N T */

extern pascal void AddHandle(Handle h, HHandleList toList);

extern pascal void AddAllRsrc(ResType rType, HHandleList toList);

extern pascal void RemHandle(Handle h, HHandleList toList);

extern pascal void ScanHandles(pascal void (*DoToHandle)(Handle h, void *DoToHandle_StaticLink), 
   void *DoToHandle_StaticLink);

		/* M E M O R Y M A N A G E M E N T */

extern pascal void BuildAllReserves(void);

extern pascal Boolean CheckReserve(void);

#if  qDebug

extern pascal void CheckRsrcUsage(void);
#endif
#if  qDebug

extern pascal void DoChangeReserve(Boolean alter, long *codeReserve, long *codeShort, long *
   lowSpaceReserve, Boolean *gotCode, Boolean *gotLowSpace);
#endif

extern pascal void FailNoReserve(void);

extern pascal void FailSpaceIsLow(void);

extern pascal void GetReserveSize(Size *szCodeReserve, Size *szMemReserve);

extern pascal Boolean MemSpaceIsLow(void);

extern pascal Ptr NewPermPtr(Size logicalSize);

extern pascal Handle NewPermHandle(Size logicalSize);

extern pascal Boolean PermAllocation(Boolean permanent);

extern pascal void SetPermHandleSize(Handle h, Size newSize);

extern pascal void SetPermPtrSize(Ptr p, Size newSize);

extern pascal void SetReserveSize(Size forCode, Size forOther);

extern pascal Size TotalTempSize(Boolean justLocked, Handle *canPurge);

#if  qDebug

extern pascal void WriteReserves(void);
#endif

		/* U T I L I T I E S */

extern pascal long AddSegSizes(Handle segRsrc);

extern pascal void SetStackSpace(long numBytes);
#endif

