// Copyright  1988, 1989 by Apple Computer, Inc.  All rights reserved.

// Includes

#include "UIconEdit.h"

// Defines

#define NIL 0											// This should be in a system or MacApp header.

// Constants

const Integer kIconHBits =			32;					// Number of horizontal bits in a bitmap.
const Integer kIconVBits =			32;					// Number of vertical bits in a bitmap.

const Integer kIconSizeInBytes =	128;				// Number of bytes in an bitmap.
const Integer kIconSizeInLongs =	32;					// Number of long words in a bitmap.
const Integer kMaxLong =			31;					// Max. addressable long word in bitmap.
	
// Resource identifiers

const Integer kSeedIconId =			1000;				// Id of the seed icon resource.
const Integer kIconWindowId =		1000;				// Id of the icon window 'view' resource.
const Integer kIconEditViewId =		1001;				// Id of the TIconEditView resource.
const Integer kPencilCursorId =		1000;				// Id of the pencil cursor resource.
	
// Constants for TIconEditView

const Integer kDefaultMagnification =	7;				// Default icon magnification.
const Integer kBorder =					5;				// Border in which to inset drawing.
const ResType kIconClipType =			'ICON';			// Our private clipboard data type.

// Command Numbers

const CmdNumber cZoomIn =			1000;				// Zoom In menu command.
const CmdNumber cZoomOut =			1001;				// Zoom Out menu command.
const CmdNumber cInvert =			1002;				// Invert menu command.
const CmdNumber cDrawCommand =		1003;				// For drawing in the icon edit view.

// Global Variables

Longint			NullStaticLink;							// Use &NullStaticLink when a static link

/*-----------------------------------------------------------------------------------------*/

void IconBitToByteBit (Point iconBit, Integer* byte, Integer* bit)
	/* This converts the given icon bit to a byte and bit in an array of long words. */
{
	Longint			bitNumber;

	bitNumber = iconBit.v * kIconVBits + iconBit.h;
	*byte = bitNumber / 8;
	*bit = 7 - (bitNumber % 8);
}

/*-----------------------------------------------------------------------------------------*/
#pragma segment AInit

pascal void TIconEditApplication::IIconEditApplication(OSType itsMainFileType)
{
	TIconEditView*		anIconEditView;

	IApplication(itsMainFileType);

	/* if (gCreateWithTemplates) ??? */					// Make sure linker doesn't strip these
		anIconEditView = new TIconEditView;
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AOpen

pascal struct TDocument* TIconEditApplication::DoMakeDocument(CmdNumber itsCmdNumber)
{
	TIconDocument*		anIconDocument;

	anIconDocument = new TIconDocument;					// Create a TIconDocument object.
	FailNIL(anIconDocument);							// Make sure we were successful.
	anIconDocument->IIconDocument();					// Initialize it.

	return anIconDocument;								// Return a reference to the document.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AClipboard

pascal void HdlMakeViewForAlienClipboard (OSErr error, Longint message, void* clipDocument)
{
	FreeObject((TIconDocument* )clipDocument);			// Free clipDocument if non-NIL.
};

pascal TView* TIconEditApplication::MakeViewForAlienClipboard()
{
	Longint				offset;
	TIconDocument*		clipDocument;
	FailInfo			fi;

	// Does the scrap contain my type?
	if (GetScrap(NIL, kIconClipType, &offset) > 0)
		{
		clipDocument =									// Need a document to represent the data
			(TIconDocument* )DoMakeDocument(cNoCommand); // to be put on the clipboard.

		CatchFailures(&fi, HdlMakeViewForAlienClipboard, // Install failure handler.
					  clipDocument);

		clipDocument->DoMakeViews(TRUE);				// Make a view without a window.
		clipDocument->fIconBitMap->LoadFromScrap();		// Get 'ICON' data from scrap.

		Success(&fi);									// Don't need failure handler anymore.

		return clipDocument->fIconDocView;				// Return view for Clipboard.
		}
	else
		return inherited::MakeViewForAlienClipboard();
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AOpen

pascal void HdlIIconDocument (OSErr error, Longint message, void* theDocument)
{
	FreeObject((TIconDocument* )theDocument);			// Free clipDocument if non-NIL.
};

pascal void TIconDocument::IIconDocument()
{
	TIconBitMap*		anIconBitMap;
	FailInfo			fi;

	fIconBitMap = NIL;									// Set this to NIL so that if IDocument
														// fails, TIconDocument.Free works okay.

	IDocument(kFileType, 								// This document's file type.
			  kSignature, 								// This document's creator.
			  kUsesDataFork, 							// This document does use the data fork
			  ! kUsesRsrcFork,							// but doesn't use the resource fork.
			  ! kDataOpen,								// We don't want the data fork kept open
			  ! kRsrcOpen);								// nor the resource fork.

	CatchFailures(&fi, HdlIIconDocument, this);			// Install failure handler in case we can't
														// create the bit map object.

	anIconBitMap = new TIconBitMap;						// Allocate a new icon bitmap.
	FailNIL(anIconBitMap);								// Fail if we can't allocate the handle.
	anIconBitMap->IIconBitMap();						// Initialize it.

	Success(&fi);										// Don't need failure handler anymore.

	fIconBitMap = anIconBitMap;
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AClose

pascal void TIconDocument::Free()
{
	FreeObject(fIconBitMap);							// Dispose of the bitmap if non-Nil.
	inherited::Free();
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AOpen

pascal void TIconDocument::DoInitialState()

	/*  This method is called to set the document's data to the "new" state, as when the user
		chooses to open a new document instead of an existing one.  We set the value of the
		document's icon bit map to that of a "seed" icon in our resource file.  That way we
		can the document's initial state simply by changing the "seed" icon resource. */

{
	Handle			seedIcon;

	seedIcon = GetIcon(1000 /*kSeedIconId*/);			// Get the seed icon resource.
	
	if (seedIcon != NIL)								// If we got the seed icon resource.
		fIconBitMap->SetIconBitMap(seedIcon);
#if qDebug
	else
		ProgramBreak("\pUnable to get the seed icon resource.");
#endif
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AOpen

pascal void TIconDocument::DoMakeViews (Boolean forPrinting)
{
	TWindow*			aWindow;
	TStdPrintHandler*	aPrintHandler;

	if (forPrinting)									// If for printing, only need the view.
		fIconDocView = (TIconEditView* )DoCreateViews(this, NIL, kIconEditViewId, &gZeroVPt);
	else
		{												// Otherwise need view and window.
		aWindow = NewTemplateWindow(kIconWindowId, this);
		fIconDocView = (TIconEditView* )(aWindow->FindSubView('ICON'));	// Get reference to the view.
		};

	aPrintHandler = new TStdPrintHandler;				// Create a print handler.
	FailNIL(aPrintHandler);							// Fail if unsucessful.
	aPrintHandler->IStdPrintHandler(this,				// The document.
		fIconDocView,									// The view to be printed.
		FALSE,											// Don't want square dots.
		TRUE,											// Horzontal page size is fixed.
		TRUE);											// Vertical page size is fixed.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal struct TCommand* TIconDocument::DoMenuCommand (CmdNumber aCmdNumber)
{
	TInvertCommand*			anInvertCommand;

	switch (aCmdNumber)
		{
		case cInvert:
			anInvertCommand = new TInvertCommand;
			FailNIL(anInvertCommand);
			anInvertCommand->IInvertCommand(this);
			return anInvertCommand;
		default:
			return inherited::DoMenuCommand(aCmdNumber);
		};
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AWriteFile

pascal void TIconDocument::DoNeedDiskSpace (Longint* dataForkBytes, Longint* rsrcForkBytes)
{
	inherited::DoNeedDiskSpace(dataForkBytes,			// In case parent class saves data.
							   rsrcForkBytes);

	dataForkBytes = dataForkBytes +						// Add the size of the icon in bytes to
					kIconSizeInBytes;					// the number of bytes in the data fork.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AReadFile

pascal void TIconDocument::DoRead (Integer aRefNum, Boolean rsrcExists, Boolean forPrinting)
{
	inherited::DoRead(aRefNum, rsrcExists,forPrinting);	// In case parent class reads data.
					 
	fIconBitMap->ReadFrom(aRefNum);						// Read Data from the file system into
														// the icon and handle any errors.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconDocument::DoSetupMenus()
{
	inherited::DoSetupMenus();							// Set up inherited menus.

	Enable(cInvert, TRUE);								// The icon can always be inverted.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AWriteFile

pascal void TIconDocument::DoWrite (Integer aRefNum, Boolean makingCopy)
{
	inherited::DoWrite(aRefNum, makingCopy);			// In case parent class writes data.

	fIconBitMap->WriteTo(aRefNum);						// Write data the icon to the file
														// and handle any errors.
};

/*-----------------------------------------------------------------------------------------*/
#if qDebug
#pragma segment AFields

pascal void 
TIconDocument::Fields(pascal void (*DoToField)(StringPtr fieldName, Ptr fieldAddr, 
	short fieldType, void* DoToField_StaticLink), void* DoToField_StaticLink)
{
	(*DoToField)("\pTIconDocument", NIL, bClass, DoToField_StaticLink);
	(*DoToField)("\pfIconBitMap", (Ptr) &fIconBitMap, bObject, DoToField_StaticLink);
	(*DoToField)("\pfIconDocView", (Ptr) &fIconDocView, bObject, DoToField_StaticLink);

	inherited::Fields(DoToField, DoToField_StaticLink);
};
#endif

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void InvalidateView (TView* aView, void* InvalidateView_StaticLink)
{
	aView->ForceRedraw();								// Cause the view to be redrawn.
}

pascal void TIconDocument::RedrawViews()
{
	ForAllViewsDo(InvalidateView, &NullStaticLink);		// Invalidate each of the doc's views.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDocument::InvertIcon()
{
	fIconBitMap->Invert();								// Invert the bits of the icon.
	RedrawViews();										// Make sure all views get redrawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDocument::ClearIcon()
{
	fIconBitMap->Clear();								// Clear the icon bits.
	RedrawViews();										// Make sure all views get redrawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDocument::SetIcon (TIconBitMap* newIcon)
{
	fIconBitMap = newIcon;								// Note this doesn't copy the icon.
	RedrawViews();										// Make sure all views get redrawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AOpen

pascal void TIconEditView::IRes (TDocument* itsDocument, TView* itsSuperView, Ptr* itsParams)
{
	inherited::IRes(itsDocument, itsSuperView, itsParams);

	fMagnification = kDefaultMagnification;
	fIconDocument = (TIconDocument* )itsDocument;
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ANonRes

pascal void TIconEditView::CalcMinSize (VPoint* minSize)
{
	SetVPt(minSize, kIconHBits * fMagnification + kBorder + kBorder,
					kIconVBits * fMagnification + kBorder + kBorder);
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AClipboard

pascal Boolean TIconEditView::ContainsClipType(ResType aType)
{
    return (aType == kIconClipType);
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal TCommand* TIconEditView::DoMenuCommand (CmdNumber aCmdNumber)
{
	TIconEditCommand*	anIconEditCommand;
	TIconPasteCommand*	anIconPasteCommand;

	switch (aCmdNumber)									// Decide if this command is ours
		{
		case cZoomIn:
			SetMagnification(fMagnification + 2);		// Increase size of each bit by 2 pixels.
			return gNoChanges;

		case cZoomOut:
			SetMagnification(fMagnification - 2);		// Decrease size of each bit by 2 pixels.
			return gNoChanges;

		case cCut:
		case cCopy:
		case cClear:
			{											// Return a TIconEditCommand object.
			anIconEditCommand = new TIconEditCommand;
			FailNIL(anIconEditCommand);
			anIconEditCommand->IIconEditCommand(aCmdNumber, this);
			return anIconEditCommand;
			};

		case cPaste:
			{											// Return a TIconPasteCommand object.
			anIconPasteCommand = new TIconPasteCommand;
			FailNIL(anIconPasteCommand);
			anIconPasteCommand->IIconPasteCommand(this);
			return anIconPasteCommand;
			};

		default:										// Otherwise, let someone else handle it.
			return inherited::DoMenuCommand(aCmdNumber);
		};
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal TCommand* TIconEditView::DoKeyCommand (short ch, Integer aKeyCode, EventInfo* info)
{
	TIconEditCommand*	anIconEditCommand;

	if (ch == chBackspace)
		{												// Return a TIconEditCommand object.
		anIconEditCommand = new TIconEditCommand;
		FailNIL(anIconEditCommand);
		anIconEditCommand->IIconEditCommand(cClear, this);	// Same as Clear menu command.
		return anIconEditCommand;
		};
	return inherited::DoKeyCommand(ch, aKeyCode, info);	// In case someone else wants it.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal TCommand* TIconEditView::DoMouseCommand (Point* theMouse, EventInfo* info, Point* hysteresis)
{
	Point				iconBit;
	TIconDrawCommand*	anIconDrawCommand;

	if (PointToBit(*theMouse, &iconBit))				// If point is within the icon
		{
		anIconDrawCommand = new TIconDrawCommand;		// then make a drawing command.
		FailNIL(anIconDrawCommand);
		anIconDrawCommand->IIconDrawCommand(this);
		return anIconDrawCommand;
		};
	return gNoChanges;									// else take no action.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal Boolean TIconEditView::DoSetCursor (Point localPoint, RgnHandle cursorRgn)
{
	Point			iconBit;
	CursHandle		pencilCursor;

	if (PointToBit(localPoint, &iconBit))				// If the cursor is not in the borders
		{
		pencilCursor = GetCursor(kPencilCursorId);		// Attempt to get the pencil cursor.
		if (pencilCursor != NIL)						// If we got it, then proceed.
			{
			SetCursor(*pencilCursor);					// Set the cursor on the screen.
			SetRectRgn(cursorRgn, kBorder, kBorder,		// Set the cursor region.
					   fSize.h - kBorder, fSize.v - kBorder);
			return true;								// Tell MacApp we set the cursor.
			};
		};

	return false;										// Didn't set the cursor
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconEditView::DoSetupMenus()
{
	inherited::DoSetupMenus();							// Set up inherited menus.

	Enable(cZoomIn, TRUE);								// Can always zoom in.
	Enable(cZoomOut, fMagnification > 1);				// Can zoom out if not at smallest size.

	Enable(cCut, TRUE);									// Always enable the Cut command.
	Enable(cCopy, TRUE);								// Always enable the Copy command.
	Enable(cClear, TRUE);								// Always enable the Clear command.

	CanPaste(kIconClipType);							// This view can paste 'ICON' data.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconEditView::Draw (const Rect* area)
{
	Rect			drawingRect;

	SetRect(&drawingRect, kBorder, kBorder,
						 kBorder + (kIconHBits * fMagnification),
						 kBorder + (kIconVBits * fMagnification));

	fIconDocument->fIconBitMap->Draw(&drawingRect);	
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconEditView::DrawBit (Point theBit, Boolean turnBitOn)
{
	Rect			theRect;
	Pattern			thePattern;

	SetRect(&theRect,									// Setup the bit's rect the in edit view.
			kBorder + theBit.h * fMagnification,
			kBorder + theBit.v * fMagnification,
			kBorder + (theBit.h + 1) * fMagnification,
			kBorder + (theBit.v + 1) * fMagnification);

	(turnBitOn)	? FillRect(&theRect, qd.black) : FillRect(&theRect, qd.white);								// Decide which pattern to use.
};

/*-----------------------------------------------------------------------------------------*/
#if qDebug
#pragma segment AFields

pascal void 
TIconEditView::Fields(pascal void (*DoToField)(StringPtr fieldName, Ptr fieldAddr, 
	short fieldType, void* DoToField_StaticLink), void* DoToField_StaticLink)
{
	(*DoToField)("\pTIconEditView", NIL, bClass, DoToField_StaticLink);
	(*DoToField)("\pfMagnification", (Ptr) &fMagnification, bInteger, DoToField_StaticLink);
	(*DoToField)("\pfIconDocument", (Ptr) &fIconDocument, bObject, DoToField_StaticLink);

	inherited::Fields(DoToField, DoToField_StaticLink);
};
#endif

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal Boolean TIconEditView::PointToBit (Point thePoint, Point* iconBit)
	/* Converts the given mouse point to an icon bit. */
{
	thePoint.h = thePoint.h - kBorder;					// Account for border in the edit view.
	thePoint.v = thePoint.v - kBorder;

	// If the mouse is within the edit view's icon
	if ((thePoint.h >= 0) && (thePoint.h < kIconHBits * fMagnification) &&
	   (thePoint.v >= 0) && (thePoint.v < kIconVBits * fMagnification))
		{
		iconBit->h = thePoint.h / fMagnification;		// Convert from edit view point
		iconBit->v = thePoint.v / fMagnification;		// to icon bit.
		return true;
		}

	return false;										// Point is not within the icon.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconEditView::SetMagnification (Integer magnification)
{
	fMagnification = Max(1, magnification);				// Set the new magnification.
	AdjustSize();										// Magnification affects the view's size.
	ForceRedraw();										// Force the view to be entirely redrawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AClipboard

pascal void TIconEditView::WriteToDeskScrap()
{
	fIconDocument->fIconBitMap->WriteToScrap();			// Write the icon data to the clipboard.

	inherited::WriteToDeskScrap();						// Write 'PICT' data to the desk scrap.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal void TInvertCommand::IInvertCommand (TIconDocument* itsIconDocument)
{
	ICommand(cInvert, itsIconDocument,					// Initialize the inherited data
			 NIL, NIL);									// no view or scroller to track.

	fIconDocument = itsIconDocument;					// Save a reference to the icon document.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TInvertCommand::DoIt()
{
	fIconDocument->InvertIcon();						// Invert the document's icon bitmap.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TInvertCommand::UndoIt()
{
	fIconDocument->InvertIcon();						// Uninvert the document's icon bitmap.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TInvertCommand::RedoIt()
{
	fIconDocument->InvertIcon();						// Reinvert the document's icon bitmap.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal void TIconDrawCommand::IIconDrawCommand (TIconEditView* itsIconEditView)
{
	fIconBitMap = NIL;									// Make sure we can be freed.
	fOriginalIcon = NIL;

	ICommand(cDrawCommand,								// Initialize the command
			 itsIconEditView->fIconDocument,			// Associate it with a document.
			 itsIconEditView,							// Associate it with a view.
			 itsIconEditView->GetScroller(true));		// Associate it with a scroller
	fConstrainsMouse = true;							// Want TrackConstrain called.

	fIconEditView = itsIconEditView;					// Set some convenience fields
	fIconDocument = itsIconEditView->fIconDocument;
	fIconBitMap = fIconDocument->fIconBitMap;			// Get reference to icon being drawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDrawCommand::Free()

{
	if (fCmdDone)										// If the command is done
		FreeObject(fOriginalIcon);						// dispose of original icon.
	else												// Else if command is undone
		FreeObject(fIconBitMap);						// dispose of icon that was drawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDrawCommand::DoIt()
{
	VRect		editViewExtent;

	fIconDocument->SetIcon(fIconBitMap);				// Set the document's bit map.

	// SetIcon invalidates all views of the document, including the one we just drew in.
	// To avoid flashing, validate the edit view here so it doesn't get redrawn.

	fIconEditView->GetExtent(&editViewExtent);
	fIconEditView->ValidVRect(&editViewExtent);
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDrawCommand::RedoIt()
{
	fIconDocument->SetIcon(fIconBitMap);				// Set the icon back to the new one.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDrawCommand::TrackConstrain(const VPoint* anchorPoint, const VPoint* previousPoint,
										  	 VPoint* nextPoint)
{
	SetVPt(nextPoint,
		   Max(Min(nextPoint->h, fIconEditView->fSize.h - kBorder - 1), kBorder),
		   Max(Min(nextPoint->v, fIconEditView->fSize.v - kBorder - 1), kBorder));
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDrawCommand::TrackFeedback(const VPoint* anchorPoint, const VPoint* nextPoint,
											Boolean turnItOn, Boolean mouseDidMove)
{
	// Overridden to avoid standard feedback.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal TCommand* TIconDrawCommand::TrackMouse(TrackPhase aTrackPhase, VPoint* anchorPoint,
							VPoint* previousPoint, VPoint* nextPoint, Boolean mouseDidMove)
{
	Point			fromBit, toBit, iconBit;
	Integer			lineLength;
	float			deltaH, deltaV, h, v;
	
	if (mouseDidMove)										// If mouse moved since last time
		{
		// Convert nextPoint and previousPoint to bit locations in the icon
	    fIconEditView->PointToBit(fIconEditView->ViewToQDPt(nextPoint), &toBit);
		fIconEditView->PointToBit(fIconEditView->ViewToQDPt(previousPoint), &fromBit);

		if (aTrackPhase == trackPress)						// If this is time TrackMouse was called
			{
			fOriginalIcon = fIconBitMap->Copy(); 			// Make a copy of the original.
			fTurnBitsOn = ! fIconBitMap->GetBit(toBit);		// Turn bits on or off?
			};

		// The following sets bits in the icon from the bit at previousPoint to the bit
		// at nextPoint.  It is based on a simple line-drawing algorithm.

		lineLength = Max(1, Max(abs(toBit.h - fromBit.h), abs(toBit.v - fromBit.v)));

		deltaH = (toBit.h - fromBit.h) / lineLength;
		deltaV = (toBit.v - fromBit.v) / lineLength;

		h = fromBit.h + 0.5;
		v = fromBit.v + 0.5;
		
		for (Integer i = 0; i < lineLength; i++)
			{
			iconBit.h = h;
			iconBit.v = v;

			fIconBitMap->SetBit(iconBit, fTurnBitsOn);		// Set the bit in the icon
			fIconEditView->DrawBit(iconBit, fTurnBitsOn);	// and draw it in the edit view.

			h = h + deltaH;
			v = v + deltaV;
			};

		};

	return this;											// Return same command object.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconDrawCommand::UndoIt()
{
	fIconDocument->SetIcon(fOriginalIcon);				// Restore icon to original state.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal void TIconEditCommand::IIconEditCommand (CmdNumber itsCmdNumber, TIconEditView* itsIconEditView)

{
	fOriginalIcon = NIL;								// Make sure we can be freed.
	fIconBitMap = NIL;

	ICommand(itsCmdNumber,								// Initialize the command
			 itsIconEditView->fIconDocument,			// Associate it with a document.
			 NIL,										// Pass NIL for the view and scroller
			 NIL);										// because we don't track the mouse.

	fChangesClipboard = (itsCmdNumber != cClear);		// Cut or Copy changes the clipboard.			
    fCausesChange = (itsCmdNumber != cCopy);			// Cut or Clear changes the documnt.

	fIconEditView = itsIconEditView;					// Remember reference to the view.
	fIconDocument = itsIconEditView->fIconDocument;		// Remember reference to the document.
	fIconBitMap = fIconDocument->fIconBitMap;			// Get reference to icon being edited.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconEditCommand::Free()

{
	if (fCmdDone || (fCmdNumber == cCopy))				// If the command is done or Copy
		FreeObject(fOriginalIcon);						// dispose of original icon.
	else												// Else if command is undone
		FreeObject(fIconBitMap);						// dispose of icon that was drawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void HdlIconEditCommandDoIt (OSErr error, Longint message, void* clipDocument)
{
	FreeObject((TIconDocument* )clipDocument);			// Free clipDocument if non-NIL.
};

pascal void TIconEditCommand::DoIt()
{
	TIconDocument*		clipDocument;
	FailInfo			fi;

	if (fCmdNumber != cClear)							// If the command is Cut or Copy
		{												// copy the data to the clipboard.
		clipDocument =									// Need a document to represent the data
			(TIconDocument* )(gApplication->DoMakeDocument(cNoCommand));
														// to be put on the clipboard.
		
		CatchFailures(&fi, HdlIconEditCommandDoIt,		// Install failure handler
					  clipDocument);
		
		clipDocument->DoMakeViews(TRUE);				// Make a view without a window.
		clipDocument->fIconDocView->SetMagnification(fIconEditView->fMagnification);		
		fIconBitMap->CopyDataTo(clipDocument->fIconBitMap);// Copy data to document icon.
	
		Success(&fi);									// Don't need failure handler anymore.
														// Tell the app we have a new clipboard.
		gApplication->ClaimClipboard(clipDocument->fIconDocView);
		};										

	if (fCmdNumber != cCopy)							// Cut or Clear commands clear the icon.
		{
		fOriginalIcon = fIconBitMap->Copy();			// Make a copy of the icon for undo.
		fIconDocument->ClearIcon();						// Clear the icon.
		};
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconEditCommand::RedoIt()
{
	if (fCmdNumber != cCopy)							// Cut or Clear commands clear the icon.
		fIconDocument->SetIcon(fIconBitMap);
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconEditCommand::UndoIt()
{
	if (fCmdNumber != cCopy)							// If the command was Cut or Clear, the
		fIconDocument->SetIcon(fOriginalIcon);			// icon was cleared, so restore it.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ASelCommand

pascal void TIconPasteCommand::IIconPasteCommand (TIconEditView* itsIconEditView)
{
	fOriginalIcon = NIL;

	ICommand(cPaste,									// Initialize the command
			 itsIconEditView->fIconDocument,			// Associate it with a document.
			 NIL,										// Pass NIL for the view and scroller
			 NIL);										// because we don't track the mouse.

	fIconEditView = itsIconEditView;					// Remember reference to the view.
	fIconDocument = itsIconEditView->fIconDocument;		// Remember reference to the document.
	fOriginalIcon = fIconDocument->fIconBitMap;			// Get reference to original icon.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconPasteCommand::Free()
{
	if (fCmdDone || (fCmdNumber == cCopy))				// If the command is done or Copy
		FreeObject(fOriginalIcon);						// dispose of original icon.
	else												// Else if command is undone
		FreeObject(fClipIcon);							// dispose of icon that was drawn.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconPasteCommand::DoIt()
{
	TIconEditView*		clipView;

	clipView = (TIconEditView* )gClipView;				// Get Clipboard view as desired type.

	fClipIcon = clipView->fIconDocument->fIconBitMap->Copy(); // Get a copy of the clipboard icon.

	fIconDocument->SetIcon(fClipIcon);					// Set the new icon.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconPasteCommand::RedoIt()
{
	fIconDocument->SetIcon(fClipIcon);					// Redo the paste.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ADoCommand

pascal void TIconPasteCommand::UndoIt()
{
	fIconDocument->SetIcon(fOriginalIcon);				// Restore the original icon.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::IIconBitMap()
{
	fDataHandle = NewPermHandle(kIconSizeInBytes);		// Allocate a handle for the bitmap.
	FailNIL(fDataHandle);								// Fail if we can't allocate the handle.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::Free()
{
	DisposIfHandle(fDataHandle);						// dispose of icon data.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::SetIconBitMap(Handle theBitMap)
{
	BlockMove(*theBitMap, *fDataHandle,					// then copy it into the document's
				kIconSizeInBytes);						// icon bitmap.
};
			
/*-----------------------------------------------------------------------------------------*/
#pragma segment AReadFile

pascal void TIconBitMap::ReadFrom (Integer aRefNum)
{
	Longint				numberOfBytes;

	numberOfBytes = kIconSizeInBytes;
	FailOSErr(FSRead(aRefNum, &numberOfBytes,			// Use Mac File System to read data
					 *fDataHandle));					// directly into the bitmap handle,
														// fail if FSRead returns an error.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment AWriteFile

pascal void TIconBitMap::WriteTo (Integer aRefNum)
{
	Longint				numberOfBytes;

	numberOfBytes = kIconSizeInBytes;
	FailOSErr(FSWrite(aRefNum,	&numberOfBytes,			// Use Mac File System to write data
					 *fDataHandle));					// directly from the bitmap handle,
					 									// fail if FSWrite returns an error.
};
			
/*-----------------------------------------------------------------------------------------*/
#pragma segment AClipboard

pascal void TIconBitMap::LoadFromScrap()
{
	Longint				offset;
	Longint				err;

	err = GetScrap(fDataHandle, kIconClipType, &offset);// Load the icon's data from the
	if (err < 0) FailOSErr(err);						// the clip board and handle errors.
};
			
/*-----------------------------------------------------------------------------------------*/
#pragma segment AClipboard

pascal void TIconBitMap::WriteToScrap()
{
    FailOSErr(PutDeskScrapData(kIconClipType,			// Write the handle to the desk scrap,
							   fDataHandle));			// failing if an error occurs.
};
			
/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal TIconBitMap* TIconBitMap::Copy()
{
	TIconBitMap*	copyOfIcon;

	copyOfIcon = new TIconBitMap;						// Create a TIconBitMap object.
	FailNIL(copyOfIcon);								// Make sure we were successful.
	copyOfIcon->IIconBitMap();							// Initialize it.
	copyOfIcon->SetIconBitMap(fDataHandle);				// Copy the data.
	return copyOfIcon;									// Return a reference to the new handle.
};
			
/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::CopyDataTo (TIconBitMap* anIcon)
{
	anIcon->SetIconBitMap(fDataHandle);					// Copy data to the new icon.
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::Clear()
{
	Longint*		longPtr;

	longPtr = *fDataHandle;
	for (Integer i = 0; i < kIconSizeInLongs; i++)
		*longPtr++ = 0;
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::Invert()
{
	Longint*		longPtr;

	longPtr = *fDataHandle;
	for (Integer i = 0; i < kIconSizeInLongs; i++)
		*longPtr++ = ~*longPtr;
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::Draw (const Rect* area)
{
	PlotIcon(area, fDataHandle);	
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal Boolean TIconBitMap::GetBit (Point iconBit)
	/* Returns the state of the given bit in the icon being drawn. */
{
	Integer			byte, bitInByte;

	IconBitToByteBit(iconBit, &byte, &bitInByte);

	return (*(*fDataHandle + byte) >> bitInByte) & 1;
};

/*-----------------------------------------------------------------------------------------*/
#pragma segment ARes

pascal void TIconBitMap::SetBit (Point iconBit, Boolean turnBitOn)
	/* Set the state of the given bit in the icon being drawn. */
{
	Integer			byte, bitInByte;

	IconBitToByteBit(iconBit, &byte, &bitInByte);

	if (turnBitOn)
		*(*fDataHandle + byte) |= (1 << bitInByte);
	else
		*(*fDataHandle + byte) &= (0 << bitInByte);
};

/*-----------------------------------------------------------------------------------------*/
#if qDebug
#pragma segment AFields

pascal void 
TIconBitMap::Fields(pascal void (*DoToField)(StringPtr fieldName, Ptr fieldAddr, 
	short fieldType, void* DoToField_StaticLink), void* DoToField_StaticLink)
{
	(*DoToField)("\pTIconBitMap", NIL, bClass, DoToField_StaticLink);
	(*DoToField)("\pfDataHandle", (Ptr) &fDataHandle, bHandle, DoToField_StaticLink);

	inherited::Fields(DoToField, DoToField_StaticLink);
};
#endif

